@@ -23,8 +23,8 @@ class JAANet(nn.Module):
         """
         # self.imgs = img_data
         # self.land_data = land_data
-        super(JAANet,self).__init__()
-        
+        super(JAANet, self).__init__()
+
         self.params = {
             "config_unit_dim": 8,
             "config_crop_size": 176,
@@ -34,7 +34,7 @@ class JAANet(nn.Module):
             "config_fill_coeff": 0.56,
             "config_write_path_prefix": get_resource_path(),
         }
-        
+
         config_unit_dim = self.params["config_unit_dim"]
         config_crop_size = self.params["config_crop_size"]
         config_map_size = self.params["config_map_size"]
@@ -68,9 +68,8 @@ class JAANet(nn.Module):
         self.au_net = network.network_dict["AUNet"](
             au_num=config_au_num, input_dim=12000, unit_dim=config_unit_dim
         )
-        
-        self.use_gpu = torch.cuda.is_available()
 
+        self.use_gpu = torch.cuda.is_available()
 
         if self.use_gpu:
             self.region_learning = self.region_learning.cuda()
@@ -155,7 +154,6 @@ class JAANet(nn.Module):
         self.global_au_feat.eval()
         self.au_net.eval()
 
-
     def align_face_49pts(self, img, img_land, box_enlarge=2.9, img_size=200):
         """
         code from:
@@ -260,17 +258,15 @@ class JAANet(nn.Module):
         return aligned_img, new_land
 
     def detect_au(self, imgs, land_data):
-        
-        land_data = np.transpose(land_data)
-        if land_data.shape[-1] == 68:
-           land_data = convert68to49(land_data)
 
-        # if land_data.shape[]
-        land_data = land_data.flatten()
-        img, land = self.align_face_49pts(imgs, land_data)  # Transforms
+        lenth_index = [len(ama) for ama in land_data]
+        lenth_cumu = np.cumsum(lenth_index)
 
-        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
-        im_pil = Image.fromarray(img)  # Convert image to PIL
+        flat_faces = np.array(
+            [item for sublist in land_data for item in sublist]
+        )  # Flatten the faces
+        flat_faces = flat_faces.transpose(0, 2, 1)
+        pt49_array = None
 
         img_transforms = transforms.Compose(
             [
@@ -280,20 +276,44 @@ class JAANet(nn.Module):
             ]
         )
 
-        input = img_transforms(im_pil)
-        if len(input.shape) < 4:
-            input.unsqueeze_(0)
-        land = torch.from_numpy(land)
+        input_torch = None
+        land_torch = None
+        for i in range(flat_faces.shape[0]):
+
+            frame_assignment = np.where(i <= lenth_cumu)[0][0]  # which frame is it?
+
+            land_convert = convert68to49(flat_faces[i]).T
+            new_land_data = land_convert.flatten()
+            new_img, new_land = self.align_face_49pts(
+                imgs[frame_assignment], new_land_data
+            )
+            new_img = cv2.cvtColor(new_img, cv2.COLOR_BGR2RGB)
+            im_pil = Image.fromarray(new_img)
+            input = img_transforms(im_pil)
+            if len(input.shape) < 4:
+                input.unsqueeze_(0)
+            new_land = torch.from_numpy(new_land)
+
+            if input_torch is None:
+                input_torch = input
+            else:
+                input_torch = torch.cat((input_torch, input), 0)
+            if land_torch is None:
+                land_torch = new_land
+            else:
+                land_torch = torch.cat((land_torch, new_land), 0)
 
         if self.use_gpu:
-            input, land = input.cuda(), land.cuda()
+            input_torch, land_torch = input_torch.cuda(), land_torch.cuda()
 
-        region_feat = self.region_learning(input)
+        region_feat = self.region_learning(input_torch)
         align_feat, align_output, aus_map = self.align_net(region_feat)
         if self.use_gpu:
             aus_map = aus_map.cuda()
         output_aus_map = self.local_attention_refine(aus_map.detach())
-        local_au_out_feat, local_aus_output = self.local_au_net(region_feat, output_aus_map)
+        local_au_out_feat, local_aus_output = self.local_au_net(
+            region_feat, output_aus_map
+        )
         local_aus_output = (local_aus_output[:, 1, :]).exp()
         global_au_out_feat = self.global_au_feat(region_feat)
         concat_au_feat = torch.cat(
