@@ -1,18 +1,21 @@
-from __future__ import division
-
-"""Plotting Functions"""
+"""
+Helper functions for plotting
+"""
 
 import numpy as np
 from sklearn.cross_decomposition import PLSRegression
+from sklearn.preprocessing import PolynomialFeatures, scale
 import matplotlib.pyplot as plt
-from feat.utils import load_h5
-
-# from utils import load_h5
+from feat.utils import load_viz_model
+from feat.pretrained import AU_LANDMARK_MAP
+from math import sin, cos
 import warnings
 import seaborn as sns
 import matplotlib.colors as colors
-from collections import OrderedDict
-from sklearn.preprocessing import minmax_scale
+from sklearn.preprocessing import minmax_scale, MinMaxScaler
+from pathlib import Path
+from PIL import Image
+from textwrap import wrap
 
 __all__ = [
     "draw_lineface",
@@ -21,8 +24,10 @@ __all__ = [
     "draw_muscles",
     "get_heat",
     "predict",
+    "imshow",
+    "interpolate_aus",
+    "animate_face",
 ]
-__author__ = ["Sophie Byrne", "Jin Hyun Cheong", "Luke Chang"]
 
 
 def draw_lineface(
@@ -34,7 +39,7 @@ def draw_lineface(
     linewidth=1,
     gaze=None,
     *args,
-    **kwargs
+    **kwargs,
 ):
     """Plot Line Face
 
@@ -91,7 +96,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
 
     eye_l = plt.Line2D(
@@ -101,7 +106,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
 
     eye_r = plt.Line2D(
@@ -111,7 +116,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
 
     eyebrow_l = plt.Line2D(
@@ -121,7 +126,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
 
     eyebrow_r = plt.Line2D(
@@ -131,7 +136,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
 
     lips1 = plt.Line2D(
@@ -169,7 +174,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
 
     lips2 = plt.Line2D(
@@ -207,7 +212,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
 
     nose1 = plt.Line2D(
@@ -217,7 +222,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
 
     nose2 = plt.Line2D(
@@ -227,7 +232,7 @@ def draw_lineface(
         linestyle=linestyle,
         linewidth=linewidth,
         *args,
-        **kwargs
+        **kwargs,
     )
     if gaze is None:
         gaze = [0, 0, 0, 0]
@@ -272,6 +277,7 @@ def draw_lineface(
             scale_units="xy",
             scale=1,
         )
+    return ax
 
 
 def draw_vectorfield(
@@ -311,8 +317,9 @@ def draw_vectorfield(
         scale_units="xy",
         scale=scale,
         *args,
-        **kwargs
+        **kwargs,
     )
+    return ax
 
 
 def draw_muscles(currx, curry, au=None, ax=None, *args, **kwargs):
@@ -785,9 +792,9 @@ def draw_muscles(currx, curry, au=None, ax=None, *args, **kwargs):
         color="w",
     )
 
-    # ax.add_patch(eye_l)
-    # ax.add_patch(eye_r)
-    # ax.add_patch(mouth)
+    ax.add_patch(eye_l)
+    ax.add_patch(eye_r)
+    ax.add_patch(mouth)
     return ax
 
 
@@ -803,7 +810,7 @@ def get_heat(muscle, au, log):
     Returns:
         color of muscle according to its au value
     """
-    q = sns.color_palette("Blues", 101)
+    q = sns.color_palette("Blues", 151)
     unit = 0
     aus = {
         "masseter_l": 15,
@@ -854,7 +861,7 @@ def get_heat(muscle, au, log):
     else:
         num = int(au[unit])
     # set alpha (opacity)
-    alpha = au[unit]/100
+    alpha = au[unit] / 100
     # color = colors.to_hex(q[num])
     # return str(color)
     color = colors.to_rgba(q[num], alpha=alpha)
@@ -872,13 +879,15 @@ def plot_face(
     linewidth=1,
     linestyle="-",
     gaze=None,
+    muscle_scaler=None,
     *args,
-    **kwargs
+    **kwargs,
 ):
-    """Function to plot facesself
+    """Core face plotting function
 
     Args:
-        model: sklearn PLSRegression instance
+        model: (str/PLSRegression instance) Name of AU visualization model to use.
+        Default's to Py-Feat's 20 AU landmark AU model
         au: vector of action units (same length as model.n_components)
         vectorfield: (dict) {'target':target_array,'reference':reference_array}
         muscles: (dict) {'muscle': color}
@@ -893,8 +902,8 @@ def plot_face(
         ax: plot handle
     """
 
-    if model is None:
-        model = load_h5()
+    if model is None or isinstance(model, str):
+        model = load_viz_model(model)
     else:
         if not isinstance(model, PLSRegression):
             raise ValueError("make sure that model is a PLSRegression instance")
@@ -902,22 +911,25 @@ def plot_face(
     if au is None:
         au = np.zeros(model.n_components)
         warnings.warn(
-            "Don't forget to pass an 'au' vector of len(20), "
+            f"Don't forget to pass an 'au' vector of length 20, "
             "using neutral as default"
         )
-        
-    landmarks = predict(au, model, feature_range = feature_range)
+
+    landmarks = predict(au, model, feature_range=feature_range)
     currx, curry = [landmarks[x, :] for x in range(2)]
 
     if ax is None:
         ax = _create_empty_figure()
 
     if muscles is not None:
-        # Muscles are always scaled 0 - 100 b/c color palette is 0-100
-        au = minmax_scale(au, feature_range=(0,100))
         if not isinstance(muscles, dict):
             raise ValueError("muscles must be a dictionary ")
-        draw_muscles(currx, curry, ax=ax, au=au, **muscles)
+        if muscle_scaler is None:
+            # Muscles are always scaled 0 - 100 b/c color palette is 0-100
+            au = minmax_scale(au, feature_range=(0, 100))
+        else:
+            au = muscle_scaler.transform(np.array(au).reshape(-1, 1)).squeeze()
+        ax = draw_muscles(currx, curry, ax=ax, au=au, **muscles)
 
     if gaze is not None and len((gaze)) != 4:
         warnings.warn(
@@ -926,7 +938,8 @@ def plot_face(
         )
         gaze = None
 
-    draw_lineface(
+    title = kwargs.pop("title", None)
+    ax = draw_lineface(
         currx,
         curry,
         color=color,
@@ -935,7 +948,7 @@ def plot_face(
         ax=ax,
         gaze=gaze,
         *args,
-        **kwargs
+        **kwargs,
     )
     if vectorfield is not None:
         if not isinstance(vectorfield, dict):
@@ -944,11 +957,18 @@ def plot_face(
             raise ValueError("vectorfield must contain 'reference' key")
         if "target" not in vectorfield.keys():
             vectorfield["target"] = landmarks
-        draw_vectorfield(ax=ax, **vectorfield)
+        ax = draw_vectorfield(ax=ax, **vectorfield)
     ax.set_xlim([25, 172])
     ax.set_ylim((240, 50))
     ax.axes.get_xaxis().set_visible(False)
     ax.axes.get_yaxis().set_visible(False)
+    if title is not None:
+        _ = ax.set_title(
+            "\n".join(wrap(title)),
+            loc="left",
+            wrap=True,
+            fontsize=14,
+        )
     return ax
 
 
@@ -964,14 +984,14 @@ def predict(au, model=None, feature_range=None):
         landmarks: Array of landmarks (2,68)
     """
     if model is None:
-        model = load_h5()
+        model = load_viz_model()
     elif not isinstance(model, PLSRegression):
         raise ValueError("make sure that model is a PLSRegression instance")
 
     if len(au) != model.n_components:
         print(au)
         print(model.n_components)
-        raise ValueError("au vector must be len(", model.n_components, ").")
+        raise ValueError(f"au vector must be length {model.n_components}.")
 
     if len(au.shape) == 1:
         au = np.reshape(au, (1, -1))
@@ -980,16 +1000,247 @@ def predict(au, model=None, feature_range=None):
         au = minmax_scale(au, feature_range=feature_range, axis=1)
 
     landmarks = np.reshape(model.predict(au), (2, 68))
-    # landmarks[1, :] = -1 * landmarks[1, :]  # this might not generalize to other models
     return landmarks
 
 
-def _create_empty_figure(figsize=(4, 5), xlim=[25, 172], ylim=[240, 50]):
+def draw_facepose(pose, facebox, ax):
+    """
+    Draw the face pose axes on the passed image for the passed facebox.
+    Adapted from draw_axis function at:
+    https://github.com/natanielruiz/deep-head-pose/blob/master/code/utils.py
+    Args:
+      img: a PIL image
+      pose: [pitch, roll, yaw] array
+      facebox: [x1, x2, width, height] array
+      ax = pyplot axis to draw on
+    """
+
+    # Center axis on facebox
+    x1, y1, w, h = facebox[:4]
+    x2, y2 = x1 + w, y1 + h
+    tdx = (x1 + x2) / 2
+    tdy = (y1 + y2) / 2
+
+    # Make rotation axis lines proportional to facebox size
+    size = min(x2 - x1, y2 - y1) // 2
+
+    # Get pose axes
+    pitch, roll, yaw = pose
+    pitch = pitch * np.pi / 180
+    yaw = -(yaw * np.pi / 180)
+    roll = roll * np.pi / 180
+
+    # X-Axis pointing to right. drawn in red
+    x1 = size * (cos(yaw) * cos(roll)) + tdx
+    y1 = size * (cos(pitch) * sin(roll) + cos(roll) * sin(pitch) * sin(yaw)) + tdy
+
+    # Y-Axis | drawn in green
+    x2 = size * (-cos(yaw) * sin(roll)) + tdx
+    y2 = size * (cos(pitch) * cos(roll) - sin(pitch) * sin(yaw) * sin(roll)) + tdy
+
+    # Z-Axis (out of the screen) drawn in blue
+    x3 = size * (sin(yaw)) + tdx
+    y3 = size * (-cos(yaw) * sin(pitch)) + tdy
+
+    # Draw face and pose axes
+    # ax.imshow(img)
+    ax.plot((tdx, x1), (tdy, y1), color="red", linewidth=2)
+    ax.plot((tdx, x2), (tdy, y2), color="green", linewidth=2)
+    ax.plot((tdx, x3), (tdy, y3), color="blue", linewidth=2)
+
+    return ax
+
+
+def _create_empty_figure(
+    figsize=(4, 5), xlim=[25, 172], ylim=[240, 50], return_fig=False
+):
     """Create an empty figure"""
-    plt.figure(figsize=figsize)
+    fig = plt.figure(figsize=figsize)
     ax = plt.gca()
     ax.set_xlim(xlim)
     ax.set_ylim(ylim)
     ax.axes.get_xaxis().set_visible(False)
     ax.axes.get_yaxis().set_visible(False)
+    if return_fig:
+        return ax, fig
     return ax
+
+
+def imshow(obj, figsize=None, aspect="equal"):
+    """
+    Convenience wrapper function around matplotlib imshow that creates figure and axis
+    boilerplate for single image plotting
+
+    Args:
+        obj (str/Path/PIL.Imag): string or Path to image file or pre-loaded PIL.Image instance
+        figsize (tuple, optional): matplotlib figure size. Defaults to None.
+        aspect (str, optional): passed to matplotlib imshow. Defaults to "equal".
+    """
+    if isinstance(obj, (str, Path)):
+        obj = Image.open(obj)
+
+    _, ax = plt.subplots(figsize=figsize)
+    _ = ax.imshow(obj, aspect=aspect)
+    _ = ax.axis("off")
+
+
+def interpolate_aus(
+    start,
+    end,
+    num_frames,
+    interp_func=None,
+    num_padding_frames=None,
+    include_reverse=True,
+):
+    """
+    Helper function to interpolate between starting and ending AU values using
+    non-linear easing functions
+
+    Args:
+        start (np.ndarray): array of starting intensities
+        end (np.ndarray): array of ending intensities
+        num_frames (int): number of frames to interpolate over
+        interp_func (callable, optional): easing function. Defaults to None.
+        num_padding_frames (int, optional): number of additional freeze frames to add
+        before the first frame and after the last frame. Defaults to None.
+        include_reverse (bool, optional): return the reverse interpolation appended to
+        the end of the interpolation. Useful for animating start -> end -> start. Defaults to True.
+
+    Returns:
+        np.ndarray: frames x au 2d array
+    """
+
+    from easing_functions import CubicEaseInOut
+
+    func = CubicEaseInOut if interp_func is None else func
+    # Loop over each AU and generate a cubic bezier style interpolation from its
+    # starting intensity to its ending intensity
+    au_interpolations = []
+    for au_start, au_end in zip(start, end):
+        interp_func = func(au_start, au_end)
+        intensities = [*map(interp_func, np.linspace(0, 1, num_frames))]
+        au_interpolations.append(intensities)
+
+    au_interpolations = np.column_stack(au_interpolations)
+
+    if num_padding_frames is not None:
+        begin_padding = np.tile(au_interpolations[0], (num_padding_frames, 1))
+        end_padding = np.tile(au_interpolations[-1], (num_padding_frames, 1))
+        au_interpolations = np.vstack([begin_padding, au_interpolations, end_padding])
+
+    if include_reverse:
+        au_interpolations = np.vstack([au_interpolations, au_interpolations[::-1, :]])
+    return au_interpolations
+
+
+def animate_face(
+    AU=None, start=None, end=None, save=None, include_reverse=True, **kwargs
+):
+    """
+    Create a matplotlib animation interpolating between a starting and ending face. Can
+    either work like `plot_face` by taking an array of AU intensities for `start` and
+    `end`, or by animating a single AU using the `AU` keyword argument and setting
+    `start` and `end` to a scalar value.
+
+    Args:
+        AU (str/int, optional): action unit id (e.g. 12 or 'AU12'). Defaults to None.
+        start (float/np.ndarray, optional): AU intensity to start at. Defaults to None.
+        end (float/np.ndarray, optional): AU intensity(s) to end at. We don't recommend
+        going beyond 3. Defaults to None.
+        save (str, optional): file to save animation to. Defaults to None.
+        include_reverse (bool, optional): Whether to also reverse the animation, i.e.
+        start -> end -> start. Defaults to True.
+        title (str, optional): plot title. Defaults to None.
+        fps (int, optional): frame-rate; Defaults to 15fps
+        duration (float, optional): length of animation in seconds. Defaults to 0.5
+        padding (float, optional): additional time to wait in seconds on the first and
+        last frame of the animation. Useful when you plan to loop the animation.
+        Defaults to 0.25
+        interp_func (callable, optional): interpolation function that takes a start and
+        end keyword argument and returns a function that will be applied to values
+        np.linspace(0, 1, num_frames); Defaults to CubicEaseInOut. See
+        https://github.com/semitable/easing-functions for other options.
+
+
+    Returns:
+        matplotlib Animation
+    """
+    from celluloid import Camera
+
+    color = kwargs.pop("color", "k")
+    linewidth = kwargs.pop("linewidth", 1)
+    linestyle = kwargs.pop("linestyle", "-")
+    fps = kwargs.pop("fps", 15)
+    duration = kwargs.pop("duration", 0.5)
+    padding = kwargs.pop("padding", 0.25)
+    num_frames = int(np.ceil(fps * duration))
+    interp_func = kwargs.pop("interp_func", None)
+
+    if AU is not None:
+        if not isinstance(start, (int, float)) or not isinstance(end, (int, float)):
+            raise TypeError("If AU is specified start and end should be single numbers")
+
+        if isinstance(AU, int):
+            AU = f"AU{str(AU).zfill(2)}"
+        au_map = dict(zip(AU_LANDMARK_MAP["Feat"], list(range(20))))
+        au_idx = au_map[AU.upper()]
+        _start, _end = np.zeros(20), np.zeros(20)
+        _start[au_idx] = start
+        _end[au_idx] = end
+        start, end = _start, _end
+
+    # To properly animate muscles we need to min-max scale referenced against the ending
+    # AU intensities rather than the AUs of any given frame of the animation which is
+    # what plot_face does if muscle_scaler is None
+    if "muscles" in kwargs:
+        muscle_scaler = MinMaxScaler(feature_range=(0, 100))
+        # MinMaxScaler defaults to per-feature normalization so reshape like we have
+        # multiple observations of a single feature
+        _ = muscle_scaler.fit(np.array(end).reshape(-1, 1))
+    else:
+        muscle_scaler = None
+
+    # Loop over each AU and generate a cubic bezier style interpolation from its
+    # starting intensity to its ending intensity
+    num_padding_frames = padding if padding is None else int(np.ceil(fps * padding))
+    au_interpolations = interpolate_aus(
+        start,
+        end,
+        interp_func=interp_func,
+        num_frames=num_frames,
+        num_padding_frames=num_padding_frames,
+        include_reverse=include_reverse,
+    )
+    gaze_start = kwargs.pop("gaze_start", np.array([0, 0, 0, 0]))
+    gaze_end = kwargs.pop("gaze_end", np.array([0, 0, 0, 0]))
+    gaze_interpolations = interpolate_aus(
+        gaze_start,
+        gaze_end,
+        interp_func=interp_func,
+        num_frames=num_frames,
+        num_padding_frames=num_padding_frames,
+        include_reverse=include_reverse,
+    )
+
+    ax, fig = _create_empty_figure(return_fig=True)
+    camera = Camera(fig)
+
+    for aus, gaze in zip(au_interpolations, gaze_interpolations):
+        ax = plot_face(
+            model=None,
+            ax=ax,
+            au=aus,
+            gaze=gaze,
+            color=color,
+            linewidth=linewidth,
+            linestyle=linestyle,
+            muscle_scaler=muscle_scaler,
+            **kwargs,
+        )
+        # if title is not None:
+        #     _ = ax.set(title=title)
+        _ = camera.snap()
+    animation = camera.animate()
+    if save is not None:
+        animation.save(save, fps=fps)
+    return animation
