@@ -1,33 +1,22 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
-
-"""Tests for `feat` package."""
-
 import pytest
-import pandas as pd
-from pandas import DataFrame, Series
 import numpy as np
-from os.path import join, exists
-from .utils import get_test_data_path
+from os.path import join
 from feat.utils import (
     read_feat,
-    read_facet,
     read_openface,
-    read_affectiva,
     registration,
-    neutral,
     softmax,
-    load_h5,
+    load_viz_model,
+    get_test_data_path,
 )
 from feat import Fex
-from nltools.data import Adjacency
-import unittest
 
 
 def test_read_feat():
-    fex = read_feat(join(get_test_data_path(), "output.csv"))
+    fex = read_feat(join(get_test_data_path(), "Feat_Test.csv"))
     assert type(fex) == Fex
 
+
 def test_utils():
     sample = read_openface(join(get_test_data_path(), "OpenFace_Test.csv"))
     lm_cols = ["x_" + str(i) for i in range(0, 68)] + [
@@ -52,4 +41,4 @@ def test_utils():
     assert softmax(0) == 0.5
     # Test badfile.
     with pytest.raises(Exception):
-        load_h5("badfile.h5")
\ No newline at end of file
+        load_viz_model("badfile")
