@@ -1,17 +1,28 @@
-from __future__ import division
-
 """
-    FEAT Utils Class
-    ==========================================
-    read_facet: read in iMotions-FACET formatted files
-    read_affdex: read in iMotions-affdex formatted files
-    read_affectiva: read in affectiva-api formatted files
-    read_openface: read in openface formatted files
-
+Feat utility and helper functions. Some of these functions can be used directly, while
+other underly the functionality of various Detector or Fex methods.
 """
 
+import os
+import math
+import h5py
+import sys
+import contextlib
+from sklearn.cross_decomposition import PLSRegression
+from joblib import load
+from sklearn import __version__ as skversion
+import numpy as np
+import pandas as pd
+from scipy.integrate import simps
+import feat
+import cv2
+import math
+import torch
+from torchvision.datasets.utils import download_url as tv_download_url
+
 __all__ = [
     "get_resource_path",
+    "get_test_data_path",
     "read_facet",
     "read_affdex",
     "read_affectiva",
@@ -19,25 +30,11 @@ __all__ = [
     "softmax",
     "registration",
     "neutral",
-    "load_h5",
+    "load_viz_model",
+    "read_pictures",
+    "validate_input",
+    "download_url",
 ]
-__author__ = ["Jin Hyun Cheong, Tiankang Xie"]
-
-
-import os, math, pywt, pickle, h5py
-from sklearn.cross_decomposition import PLSRegression
-# setattr(PLSRegression, "_x_mean", None)
-# setattr(PLSRegression, "_y_mean", None)
-# setattr(PLSRegression, "_x_std", None)
-from sklearn import __version__
-import numpy as np, pandas as pd
-from scipy import signal
-from scipy.integrate import simps
-import feat
-import cv2
-import math
-from PIL import Image
-import torch
 
 """ DEFINE IMPORTANT VARIABLES """
 # FEAT columns
@@ -100,13 +97,6 @@ openface_3d_landmark_columns = (
     + [f"Z_{i}" for i in range(landmark_length)]
 )
 
-jaanet_AU_list = [1, 2, 4, 6, 7, 10, 12, 14, 15, 17, 23, 24]
-RF_AU_list = [1,2,4,5,6,7,9,10,11,12,14,15,17,20,23,24,25,26,28,43]
-jaanet_AU_presence = [f"AU" + str(i).zfill(2) for i in jaanet_AU_list]
-jaanet_AU_presence.sort()
-RF_AU_presence = [f"AU" + str(i).zfill(2) for i in RF_AU_list]
-RF_AU_presence.sort()
-
 openface_AU_list = [1, 2, 4, 5, 6, 7, 9, 10, 12, 14, 15, 17, 20, 23, 25, 26, 45]
 openface_AU_intensity = [f"AU" + str(i).zfill(2) + "_r" for i in openface_AU_list]
 openface_AU_presence = [f"AU" + str(i).zfill(2) + "_c" for i in openface_AU_list + [28]]
@@ -195,39 +185,126 @@ def face_rect_to_coords(rectangle):
 
 
 def get_resource_path():
-    """ Get path to feat resource directory. """
-    return os.path.join(feat.__path__[0], "resources")  # points to the package folder.
-    # return ("F:/feat/feat/") # points to the package folder.
-    # return os.path.join(os.path.dirname(__file__), 'resources')
+    """Get path to feat resource directory."""
+    return os.path.join(feat.__path__[0], "resources")
 
-def load_h5(file_name="pyfeat_aus_to_landmarks.h5"):
-    """Load the h5 PLS model for plotting.
+
+def get_test_data_path():
+    """Get path to feat test data directory."""
+    return os.path.join(feat.__path__[0], "tests", "data")
+
+
+def load_viz_model(
+    file_name=None,
+    prefer_joblib_if_version_match=True,
+    verbose=False,
+):
+    """Load the h5 PLS model for plotting. Will try using joblib if python and sklearn
+    major and minor versions match those the model was trained with (3.8.x and 1.0.x
+    respectively), otherwise will reconstruct the model object using h5 data.
 
     Args:
         file_name (str, optional): Specify model to load.. Defaults to 'blue.h5'.
+        prefer_joblib_if_version_match (bool, optional): If the sklearn and python major.minor versions
+        match then return the pickled PLSRegression object. Otherwise build it from
+        scratch using .h5 data. Default True
 
     Returns:
         model: PLS model
     """
+
+    file_name = "pyfeat_aus_to_landmarks" if file_name is None else file_name
+
+    if "." in file_name:
+        raise TypeError("Please use a file name with no extension")
+
+    h5_path = os.path.join(get_resource_path(), f"{file_name}.h5")
+    joblib_path = os.path.join(get_resource_path(), f"{file_name}.joblib")
+
+    # Make sure saved viz model exists
+    if not os.path.exists(h5_path):
+        raise ValueError(f"{h5_path} does not exist")
+
+    if not os.path.exists(joblib_path):
+        raise ValueError(f"{joblib_path} does not exist")
+
+    # Check sklearn and python version to see if we can load joblib
+    my_skmajor, my_skminor, my_skpatch = skversion.split(".")
+    my_pymajor, my_pyminor, my_pymicro, *_ = sys.version_info
+
+    # Versions viz models were trained with
+    pymajor, pyminor, skmajor, skminor = 3, 8, 1, 1
+    if (
+        int(my_skmajor) == skmajor
+        and int(my_skminor) == skminor
+        and int(my_pymajor) == pymajor
+        and int(my_pyminor) == pyminor
+        and prefer_joblib_if_version_match
+    ):
+        can_load_joblib = True
+    else:
+        can_load_joblib = False
+
     try:
-        hf = h5py.File(os.path.join(get_resource_path(), file_name), "r")
-        d1 = hf.get("coef")
-        d2 = hf.get("x_mean")
-        d3 = hf.get("y_mean")
-        d4 = hf.get("x_std")
-        model = PLSRegression(len(d1))
-        model.coef_ = np.array(d1)
-        if int(__version__.split(".")[1]) < 24:
-            model.x_mean_ = np.array(d2)
-            model.y_mean_ = np.array(d3)
-            model.x_std_ = np.array(d4)
+        if can_load_joblib:
+            if verbose:
+                print("Loading joblib")
+            model = load(joblib_path)
+            # We need the h5 file for some meta-data even when loading using joblib
+            hf = h5py.File(h5_path, mode="r")
+            model.__dict__["model_name_"] = hf.attrs["model_name"]
+            model.__dict__["skversion"] = hf.attrs["skversion"]
+            model.__dict__["pyversion"] = hf.attrs["pyversion"]
+            hf.close()
         else:
-            model._x_mean = np.array(d2)
-            model._y_mean = np.array(d3)
-            model._x_std = np.array(d4)
-        hf.close()
+            if verbose:
+                print("Reconstructing from h5")
+            hf = h5py.File(h5_path, mode="r")
+            x_weights = np.array(hf.get("x_weights"))
+            model = PLSRegression(n_components=x_weights.shape[1])
+            # PLSRegression in sklearn < 1.1 storex coefs as samples x features, but
+            # recent versions transpose this. Check if the user is on Python 3.7 (which
+            # only supports sklearn 1.0.x) or < sklearn 1.1.x
+            if (my_pymajor == 3 and my_pyminor == 7) or (
+                my_skmajor == 1 and my_skminor != 1
+            ):
+                model.__dict__["coef_"] = np.array(hf.get("coef"))
+                model.__dict__["_coef_"] = np.array(hf.get("coef"))
+            else:
+                model.__dict__["coef_"] = np.array(hf.get("coef")).T
+                model.__dict__["_coef_"] = np.array(hf.get("coef")).T
+            model.__dict__["x_weights_"] = np.array(hf.get("x_weights"))
+            model.__dict__["y_weights_"] = np.array(hf.get("y_weights"))
+            model.__dict__["x_loadings"] = np.array(hf.get("x_loadings"))
+            model.__dict__["y_loadings"] = np.array(hf.get("y_loadings"))
+            model.__dict__["x_scores"] = np.array(hf.get("x_scores"))
+            model.__dict__["y_scores"] = np.array(hf.get("y_scores"))
+            model.__dict__["x_rotations"] = np.array(hf.get("x_rotations"))
+            model.__dict__["y_rotations"] = np.array(hf.get("y_rotations"))
+            model.__dict__["intercept"] = np.array(hf.get("intercept"))
+            model.__dict__["x_train"] = np.array(hf.get("X_train"))
+            model.__dict__["y_train"] = np.array(hf.get("Y_train"))
+            model.__dict__["X_train"] = np.array(hf.get("x_train"))
+            model.__dict__["Y_train"] = np.array(hf.get("y_train"))
+            model.__dict__["intercept_"] = np.array(hf.get("intercept"))
+            model.__dict__["model_name_"] = hf.attrs["model_name"]
+            model.__dict__["skversion"] = hf.attrs["skversion"]
+            model.__dict__["pyversion"] = hf.attrs["pyversion"]
+
+            # Older sklearn version named these attributes differently
+            if int(skversion.split(".")[0]) < 1:
+                model.__dict__["x_mean_"] = np.array(hf.get("x_mean"))
+                model.__dict__["y_mean_"] = np.array(hf.get("y_mean"))
+                model.__dict__["x_std_"] = np.array(hf.get("x_std"))
+                model.__dict__["y_std_"] = np.array(hf.get("y_std"))
+            else:
+                model.__dict__["_x_mean"] = np.array(hf.get("x_mean"))
+                model.__dict__["_y_mean"] = np.array(hf.get("y_mean"))
+                model.__dict__["_x_std"] = np.array(hf.get("x_std"))
+                model.__dict__["_y_std"] = np.array(hf.get("y_std"))
+            hf.close()
     except Exception as e:
-        print("Unable to load data ", file_name, ":", e)
+        raise IOError(f"Unable to load data: {e}")
     return model
 
 
@@ -242,7 +319,7 @@ def read_feat(fexfile):
     """
     d = pd.read_csv(fexfile)
     au_columns = [col for col in d.columns if "AU" in col]
-    return feat.Fex(
+    fex = feat.Fex(
         d,
         filename=fexfile,
         au_columns=au_columns,
@@ -252,6 +329,8 @@ def read_feat(fexfile):
         time_columns=FEAT_TIME_COLUMNS,
         detector="Feat",
     )
+    fex["input"] = fexfile
+    return fex
 
 
 def read_facet(facetfile, features=None, raw=False, sampling_freq=None):
@@ -274,7 +353,7 @@ def read_facet(facetfile, features=None, raw=False, sampling_freq=None):
             studydate = f.readline().replace("\t", "").replace("\n", "")
             versionstr = f.readline().replace("\t", "").replace("\n", "")
         versionnum = int(versionstr.split(" ")[-1].split(".")[0])
-    except:
+    except Exception:
         raise TypeError(
             "Cannot infer version of iMotions-FACET file. Check to make sure this is the raw iMotions-FACET file."
         )
@@ -286,7 +365,7 @@ def read_facet(facetfile, features=None, raw=False, sampling_freq=None):
             d = d[features]
             if raw:
                 return feat.Fex(d, filename=facetfile)
-        except:
+        except Exception:
             raise KeyError([features, "not in facetfile"])
     elif isinstance(features, type(None)):
         if raw:
@@ -305,7 +384,7 @@ def read_facet(facetfile, features=None, raw=False, sampling_freq=None):
             d.columns = [col.replace(" ", "") for col in d.columns]
             # d._metadata = fex_columns
     au_columns = [col for col in d.columns if "AU" in col]
-    return feat.Fex(
+    fex = feat.Fex(
         d,
         filename=facetfile,
         au_columns=au_columns,
@@ -317,6 +396,8 @@ def read_facet(facetfile, features=None, raw=False, sampling_freq=None):
         detector="FACET",
         sampling_freq=sampling_freq,
     )
+    fex["input"] = facetfile
+    return fex
 
 
 def read_openface(openfacefile, features=None):
@@ -390,15 +471,15 @@ def read_openface(openfacefile, features=None):
     if isinstance(features, list):
         try:
             d = d[features]
-        except:
+        except Exception:
             raise KeyError([features, "not in openfacefile"])
     elif isinstance(features, type(None)):
         features = OPENFACE_ORIG_COLUMNS
         try:
             d = d[features]
-        except:
+        except Exception:
             pass
-    return feat.Fex(
+    fex = feat.Fex(
         d,
         filename=openfacefile,
         au_columns=openface_AU_columns,
@@ -410,6 +491,8 @@ def read_openface(openfacefile, features=None):
         time_columns=openface_time_columns,
         detector="OpenFace",
     )
+    fex["input"] = openfacefile
+    return fex
 
 
 def read_affectiva(affectivafile, orig_cols=False):
@@ -457,27 +540,30 @@ def read_affectiva(affectivafile, orig_cols=False):
         "upperLipRaise": "AU10",
     }
     affectiva_au_columns = [col for col in rep_dict.values() if "AU" in col]
+    detector = "Affectiva" if "AU01" in affectiva_au_columns else "Affectiva-named"
     affectiva_emotion_columns = list(set(rep_dict.values()) - set(affectiva_au_columns))
     if not orig_cols:
         new_cols = []
         for col in d.columns:
             try:
                 new_cols.append(rep_dict[col])
-            except:
+            except Exception:
                 new_cols.append(col)
         d.columns = new_cols
-    return feat.Fex(
+    fex = feat.Fex(
         d,
         filename=affectivafile,
         au_columns=affectiva_au_columns,
         emotion_columns=affectiva_emotion_columns,
-        detector="Affectiva",
+        detector=detector,
     )
+    fex["input"] = affectivafile
+    return fex
 
 
 def wavelet(freq, num_cyc=3, sampling_freq=30.0):
     """Create a complex Morlet wavelet.
-    
+
     Creates a complex Morlet wavelet by windowing a cosine function by a Gaussian. All formulae taken from Cohen, 2014 Chaps 12 + 13
 
     Args:
@@ -496,7 +582,7 @@ def wavelet(freq, num_cyc=3, sampling_freq=30.0):
 
     # Gaussian component
     sd = num_cyc / (2 * np.pi * freq)  # standard deviation
-    gaus = np.exp(-(time ** 2.0) / (2.0 * sd ** 2.0))
+    gaus = np.exp(-(time**2.0) / (2.0 * sd**2.0))
 
     return sin * gaus
 
@@ -507,13 +593,13 @@ def calc_hist_auc(vals, hist_range=None):
     This function follows the bag of temporal feature analysis as described in Bartlett, M. S., Littlewort, G. C., Frank, M. G., & Lee, K. (2014). Automatic decoding of facial movements reveals deceptive pain expressions. Current Biology, 24(7), 738-743. The function receives convolved data, squares the values, finds 0 crossings to calculate the AUC(area under the curve) and generates a 6 exponentially-spaced-bin histogram for each data.
 
     Args:
-        vals: 
+        vals:
 
     Returns:
         Series of histograms
     """
     # Square values
-    vals = [elem ** 2 if elem > 0 else -1 * elem ** 2 for elem in vals]
+    vals = [elem**2 if elem > 0 else -1 * elem**2 for elem in vals]
     # Get 0 crossings
     crossings = np.where(np.diff(np.sign(vals)))[0]
     pos, neg = [], []
@@ -550,6 +636,34 @@ def softmax(x):
     return 1.0 / (1 + 10.0 ** -(x))
 
 
+def validate_input(inputFname):
+    """
+    Given a string filename or list containing string files names, ensures that the
+    file(s) exist. Always returns a non-nested list, potentionally containing a single element.
+
+    Args:
+        inputFname (str or list): file name(s)
+
+    Raises:
+        FileNotFoundError: if any file name(s) don't exist
+
+    Returns:
+        list: list of file names (even if input was a str)
+    """
+
+    assert isinstance(
+        inputFname, (str, list)
+    ), "inputFname must be a string path to image or list of image paths"
+
+    if isinstance(inputFname, str):
+        inputFname = [inputFname]
+
+    for inputF in inputFname:
+        if not os.path.exists(inputF):
+            raise FileNotFoundError(f"File {inputF} not found.")
+    return inputFname
+
+
 ### Functions for face registration ###
 neutral = np.array(
     [
@@ -674,7 +788,7 @@ def convert68to49(points):
     """Convert landmark form 68 to 49
 
     Function slightly modified from https://github.com/D-X-Y/landmark-detection/blob/7bc7a5dbdbda314653124a4596f3feaf071e8589/SAN/lib/datasets/dataset_utils.py#L169 to fit pytorch tensors. Converts 68 point landmarks to 49 point landmarks
-    
+
     Args:
         points: landmark points of shape (2,68) or (3,68)
 
@@ -743,7 +857,7 @@ def drawLandmark(img, bbox, landmark):
 
     Returns:
         img marked with landmark and bbox
-    """ 
+    """
     img_ = img.copy()
     cv2.rectangle(
         img_, (bbox.left, bbox.top), (bbox.right, bbox.bottom), (0, 0, 255), 2
@@ -757,7 +871,7 @@ def drawLandmark_multiple(img, bbox, landmark):
     """Draw multiple landmarks.
 
     From https://github.com/cunjian/pytorch_face_landmark/
-    
+
     Args:
         img ([type]): gray or RGB
         bbox ([type]): type of BBox
@@ -765,12 +879,13 @@ def drawLandmark_multiple(img, bbox, landmark):
 
     Returns:
         img marked with landmark and bbox
-    """    
+    """
     cv2.rectangle(img, (bbox.left, bbox.top), (bbox.right, bbox.bottom), (0, 0, 255), 2)
     for x, y in landmark:
         cv2.circle(img, (int(x), int(y)), 2, (0, 255, 0), -1)
     return img
 
+
 def padding(img, expected_size):
     """
     DOCUMENTATION GOES HERE
@@ -780,10 +895,42 @@ def padding(img, expected_size):
     delta_height = desired_size - img.size[1]
     pad_width = delta_width // 2
     pad_height = delta_height // 2
-    padding = (pad_width, pad_height, delta_width - pad_width, delta_height - pad_height)
+    padding = (
+        pad_width,
+        pad_height,
+        delta_width - pad_width,
+        delta_height - pad_height,
+    )
     return ImageOps.expand(img, padding)
 
 
+def read_pictures(imgname_list):
+    """
+    Reads in a list of pictures and concatenate these pictures into batches of images.
+
+    Args:
+        imgname_list (list of string): a list of filenames for the facial pictures
+
+    Returns:
+        img_batch_arr (np.array): np array of shape BxHxWxC
+    """
+
+    imgname_list = validate_input(imgname_list)
+    img_batch_arr = None
+    for img_name in imgname_list:
+        frame = cv2.imread(img_name)
+        frame = np.expand_dims(frame, 0)
+        if img_batch_arr is None:
+            img_batch_arr = frame
+        else:
+            assert (
+                img_batch_arr.shape[1::] == frame.shape[1::]
+            ), "please make sure that the input images are of the same shape! otherwise you need to process each image individually"
+            img_batch_arr = np.concatenate([img_batch_arr, frame], 0)
+
+    return img_batch_arr
+
+
 def resize_with_padding(img, expected_size):
     """
     DOCUMENTATION GOES HERE
@@ -794,55 +941,132 @@ def resize_with_padding(img, expected_size):
     delta_height = expected_size[1] - img.size[1]
     pad_width = delta_width // 2
     pad_height = delta_height // 2
-    padding = (pad_width, pad_height, delta_width - pad_width, delta_height - pad_height)
+    padding = (
+        pad_width,
+        pad_height,
+        delta_width - pad_width,
+        delta_height - pad_height,
+    )
     return ImageOps.expand(img, padding)
 
 
 def align_face_68pts(img, img_land, box_enlarge, img_size=112):
     """Performs affine transformation to align the images by eyes.
 
-    Performs affine alignment including eyes. 
-    
+    Performs affine alignment including eyes.
+
     Args:
         img: gray or RGB
-        img_land: 68 system flattened landmarks, shape:(136) 
+        img_land: 68 system flattened landmarks, shape:(136)
         box_enlarge: relative size of face on the image. Smaller value indicate larger proportion
         img_size = output image size
     Return:
         aligned_img: the aligned image
-        new_land: the new landmarks 
+        new_land: the new landmarks
     """
-    leftEye0 = (img_land[2 * 36] + img_land[2 * 37] + img_land[2 * 38] + img_land[2 * 39] + img_land[2 * 40] +
-                img_land[2 * 41]) / 6.0
-    leftEye1 = (img_land[2 * 36 + 1] + img_land[2 * 37 + 1] + img_land[2 * 38 + 1] + img_land[2 * 39 + 1] +
-                img_land[2 * 40 + 1] + img_land[2 * 41 + 1]) / 6.0
-    rightEye0 = (img_land[2 * 42] + img_land[2 * 43] + img_land[2 * 44] + img_land[2 * 45] + img_land[2 * 46] +
-                img_land[2 * 47]) / 6.0
-    rightEye1 = (img_land[2 * 42 + 1] + img_land[2 * 43 + 1] + img_land[2 * 44 + 1] + img_land[2 * 45 + 1] +
-                img_land[2 * 46 + 1] + img_land[2 * 47 + 1]) / 6.0
-    deltaX = (rightEye0 - leftEye0)
-    deltaY = (rightEye1 - leftEye1)
+    leftEye0 = (
+        img_land[2 * 36]
+        + img_land[2 * 37]
+        + img_land[2 * 38]
+        + img_land[2 * 39]
+        + img_land[2 * 40]
+        + img_land[2 * 41]
+    ) / 6.0
+    leftEye1 = (
+        img_land[2 * 36 + 1]
+        + img_land[2 * 37 + 1]
+        + img_land[2 * 38 + 1]
+        + img_land[2 * 39 + 1]
+        + img_land[2 * 40 + 1]
+        + img_land[2 * 41 + 1]
+    ) / 6.0
+    rightEye0 = (
+        img_land[2 * 42]
+        + img_land[2 * 43]
+        + img_land[2 * 44]
+        + img_land[2 * 45]
+        + img_land[2 * 46]
+        + img_land[2 * 47]
+    ) / 6.0
+    rightEye1 = (
+        img_land[2 * 42 + 1]
+        + img_land[2 * 43 + 1]
+        + img_land[2 * 44 + 1]
+        + img_land[2 * 45 + 1]
+        + img_land[2 * 46 + 1]
+        + img_land[2 * 47 + 1]
+    ) / 6.0
+    deltaX = rightEye0 - leftEye0
+    deltaY = rightEye1 - leftEye1
     l = math.sqrt(deltaX * deltaX + deltaY * deltaY)
     sinVal = deltaY / l
     cosVal = deltaX / l
     mat1 = np.mat([[cosVal, sinVal, 0], [-sinVal, cosVal, 0], [0, 0, 1]])
-    mat2 = np.mat([[leftEye0, leftEye1, 1], [rightEye0, rightEye1, 1], [img_land[2 * 30], img_land[2 * 30 + 1], 1],
-                [img_land[2 * 48], img_land[2 * 48 + 1], 1], [img_land[2 * 54], img_land[2 * 54 + 1], 1]])
+    mat2 = np.mat(
+        [
+            [leftEye0, leftEye1, 1],
+            [rightEye0, rightEye1, 1],
+            [img_land[2 * 30], img_land[2 * 30 + 1], 1],
+            [img_land[2 * 48], img_land[2 * 48 + 1], 1],
+            [img_land[2 * 54], img_land[2 * 54 + 1], 1],
+        ]
+    )
     mat2 = (mat1 * mat2.T).T
     cx = float((max(mat2[:, 0]) + min(mat2[:, 0]))) * 0.5
     cy = float((max(mat2[:, 1]) + min(mat2[:, 1]))) * 0.5
-    if (float(max(mat2[:, 0]) - min(mat2[:, 0])) > float(max(mat2[:, 1]) - min(mat2[:, 1]))):
+    if float(max(mat2[:, 0]) - min(mat2[:, 0])) > float(
+        max(mat2[:, 1]) - min(mat2[:, 1])
+    ):
         halfSize = 0.5 * box_enlarge * float((max(mat2[:, 0]) - min(mat2[:, 0])))
     else:
         halfSize = 0.5 * box_enlarge * float((max(mat2[:, 1]) - min(mat2[:, 1])))
     scale = (img_size - 1) / 2.0 / halfSize
-    mat3 = np.mat([[scale, 0, scale * (halfSize - cx)], [0, scale, scale * (halfSize - cy)], [0, 0, 1]])
+    mat3 = np.mat(
+        [
+            [scale, 0, scale * (halfSize - cx)],
+            [0, scale, scale * (halfSize - cy)],
+            [0, 0, 1],
+        ]
+    )
     mat = mat3 * mat1
-    aligned_img = cv2.warpAffine(img, mat[0:2, :], (img_size, img_size), cv2.INTER_LINEAR, borderValue=(128, 128, 128))
-    land_3d = np.ones((int(len(img_land)/2), 3))
-    land_3d[:, 0:2] = np.reshape(np.array(img_land), (int(len(img_land)/2), 2))
+    aligned_img = cv2.warpAffine(
+        img,
+        mat[0:2, :],
+        (img_size, img_size),
+        cv2.INTER_LINEAR,
+        borderValue=(128, 128, 128),
+    )
+    land_3d = np.ones((int(len(img_land) / 2), 3))
+    land_3d[:, 0:2] = np.reshape(np.array(img_land), (int(len(img_land) / 2), 2))
     mat_land_3d = np.mat(land_3d)
     new_land = np.array((mat * mat_land_3d.T).T)
-    new_land = np.array(list(zip(new_land[:,0], new_land[:,1]))).astype(int)
+    new_land = np.array(list(zip(new_land[:, 0], new_land[:, 1]))).astype(int)
 
     return aligned_img, new_land
+
+
+def round_vals(list_of_arrays, ndigits):
+    list_of_arrays2 = list_of_arrays.copy()
+    for i, arr0 in enumerate(list_of_arrays):
+        for j, arr1 in enumerate(list_of_arrays):
+            list_of_arrays2[i][j] = np.around(list_of_arrays[i][j], ndigits)
+    return list_of_arrays2
+
+
+def download_url(*args, **kwargs):
+    """By default just call download_url from torch vision, but we pass a verbose =
+    False keyword argument, then call download_url with a special context manager than
+    supresses the print messages"""
+    verbose = kwargs.pop("verbose", True)
+
+    if verbose:
+        return tv_download_url(*args, **kwargs)
+
+    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
+        return tv_download_url(*args, **kwargs)
+
+
+class FaceDetectionError(Exception):
+    """Error when face detection failed in a batch"""
+
+    pass
